#include "DolibarrClient.h"
#include <WiFi.h>
#include <ArduinoJson.h>
#include <iostream>

#define WAITING_WIFI_DELAY 1000

DolibarrClient::DolibarrClient(struct WifiConfig wifi_config, struct DolibarrConfig dolibarr_config) : wifi(wifi_config), dolibarr(dolibarr_config) {
    #if defined(DEBUG)
        Serial.println(" --- Wifi configuration --- ");
        Serial.println((std::string("SSID: ") + std::string(wifi_config.ssid)).c_str());
        Serial.println((std::string("Password: ") + std::string(wifi_config.password)).c_str());
        Serial.println(" --- Dolibarr configuration --- ");
        Serial.println((std::string("Base URL: ") + std::string(dolibarr_config.url)).c_str());
        Serial.println((std::string("Token: ") + std::string(dolibarr_config.api_key)).c_str());
    #endif
    this->initialize_wifi();
    this->initialize_http_client();
}

HTTPClient *DolibarrClient::build_url(const String& url) const {
    auto *client = new HTTPClient();
    String clientUrl = this->dolibarr.url + url;
    client->begin(clientUrl);
    client->addHeader("Content-Type", "application/json");
    client->addHeader("DOLAPIKEY", this->dolibarr.api_key);
    #if defined(DEBUG)
        Serial.println("URL Request: " + clientUrl);
    #endif
    return client;
}


int DolibarrClient::login() const {
    HTTPClient *client = this->build_url(LOGIN_URL);
    int httpResponseCode = client->GET();
    if (httpResponseCode > 0) {
        StaticJsonDocument<LOGIN_JSON_SIZE> doc;
        DeserializationError error = deserializeJson(doc, client->getString().c_str());
        if (error) {
            delete client;
            return -1;
        }
        delete client;
        return 0;
    }
    delete client;
    return -1;
}

std::string replace_id(const char *str, const char *id) {
    std::string url(str);
    url.replace(url.find("{id}"), 4, id);
    return url;
}

std::vector<models::Product> *DolibarrClient::list_products() const {
    HTTPClient *client = this->build_url(LIST_PRODUCT_URL);
    if (client->GET() == OK_RESPONSE) {

    }
    return nullptr;
}

models::Product *DolibarrClient::get_product_by_id(const char* id_product) const {
    HTTPClient *client = this->build_url(replace_id(GET_PRODUCT_URL, id_product).c_str());
    if (client->GET() == OK_RESPONSE) {
        StaticJsonDocument<GET_PRODUCT_JSON_SIZE> doc;
        DeserializationError error = deserializeJson(doc, client->getString().c_str());
        if (error) {
            Serial.println("ERROR: ");
            Serial.println(error.c_str());
            delete client;
            return nullptr;
        }
        auto *product = new models::Product();
        product->date_creation = doc["date_creation"];
        product->id = doc["id"];
        product->entity = doc["entity"];
        product->stock_reel = doc["stock_reel"];
        product->label = doc["label"];
        delete client;
        return product;
    }
    delete client;
    return nullptr;
}

std::vector<models::Warehouse> *DolibarrClient::list_warehouse() const {
    HTTPClient *client = this->build_url(LIST_WAREHOUSE_URL);
    if (client->GET() == OK_RESPONSE) {
        StaticJsonDocument<LIST_WAREHOUSE_JSON_SIZE> doc;
        DeserializationError error = deserializeJson(doc, client->getString().c_str());
        if (error) {
            Serial.println("ERROR: ");
            Serial.println(error.c_str());
            delete client;
            return nullptr;
        }
        auto *warehouses = new std::vector<models::Warehouse>();
        for (auto obj : doc.as<JsonArray>()) {
            models::Warehouse warehouse = {};
            warehouse.id = obj["id"];
            warehouses->push_back(warehouse);
        }
        delete client;
        return warehouses;
    }
    delete client;
    return nullptr;
}

int DolibarrClient::create_movement(models::CreateProductStock &stock) const {
    HTTPClient *client = this->build_url(CREATE_STOCKS_MOVEMENTS_URL);
    StaticJsonDocument<CREATE_STOCKS_MOVEMENTS_JSON_SIZE> doc;
    std::string result;
    doc["product_id"] = stock.product_id;
    doc["warehouse_id"]  = stock.warehouse_id;
    doc["qty"] = stock.qty;
    serializeJson(doc, result);
    Serial.println(result.c_str());
    if (client->POST(result.c_str()) == OK_RESPONSE) {
        delete client;
        return 0;
    }
    return -1;
}

int DolibarrClient::initialize_wifi() const {
    WiFiClass::mode(WIFI_STA); //Optional
    WiFi.setSleep(false);
    WiFi.begin(this->wifi.ssid, this->wifi.password);
    Serial.print("Connecting ");
    while(WiFiClass::status() != WL_CONNECTED){
        delay(WAITING_WIFI_DELAY);
        Serial.print(".");
    }
    Serial.println("Connected to the WiFi network");
    return 0;
}

int DolibarrClient::initialize_http_client() {
    this->httpClient = new HTTPClient();
    if (this->login() == 0) {
        auto* product = this->get_product_by_id("1");
        if (product == nullptr) {
            Serial.println("Product is nullptr !");
            return -1;
        }
        Serial.println("Product label: ");
        Serial.println(product->label);
        auto* warehouses = this->list_warehouse();
        if (warehouses == nullptr) {
            delete product;
            Serial.println("Warehouse is nullptr !");
            return -1;
        }
        Serial.println("Warehouses: ");
        models::CreateProductStock product_stock = {product->id, warehouses->at(0).id, "1"};
        this->create_movement(product_stock);
        for (auto warehouse : *warehouses) {
            Serial.println(warehouse.id);
        }
        delete product;
        delete warehouses;
    } else {
        Serial.println("An Error has occurred while trying to login");
    }
    return 0;
}